#!/usr/bin/python3

'''
Written by Martin Olivesten in 2023 for TDDD55.
Full permission to use in anyway with no restrictions. :)

Run: `python3 test.py -h` for usage.
'''

from subprocess import Popen, PIPE, STDOUT
from re import search
from argparse import ArgumentParser as ArgParse
from dataclasses import dataclass
from random import randrange
from io import TextIOWrapper
from typing import List, Tuple

GREEN = '\x1b[32m'
YELLOW = '\x1b[33m'
RED = '\x1b[31m'
RESET = '\x1b[39m'

PASS = f'{GREEN}PASS{RESET}'
FAIL = f'{RED}FAIL{RESET}'
MAYBE = f'{YELLOW}MAYBE{RESET}'

REAL_REGEX = r'-?(\.[0-9]+|[0-9]+(\.[0-9]*)?)([eE][+-]?[0-9]+)?'

dump = ''


@dataclass
class Result():
    expr: str
    expected: str
    got: str
    outcome: str


def parse_tests(file: TextIOWrapper) -> List[Tuple[str, str]]:
    tests = file.readlines()
    tests = filter(lambda t: t[0] != '#' and t.strip(), tests)
    tests = map(lambda t: t.split('='), tests)
    tests = list(tests)
    for t in tests:
        if len(t) != 2:
            print(f'ERR: Invalid test format: `{", ".join(t).strip()}`.')
            exit(1)

    tests = map(lambda p: list(map(str.strip, p)), tests)
    return list(tests)


def run_tests(tests: List[Tuple[str, str]], prog: str) -> (List[Result], int):
    global dump
    res = []
    n_err = 0
    for i, test in enumerate(tests):
        expr, expected = test
        p = Popen([prog], stdin=PIPE, stdout=PIPE, stderr=STDOUT)
        p_out = p.communicate(f'{expr}\n'.encode())[0].decode()
        m = search(fr'Result:\s*{REAL_REGEX}', p_out)

        if m is None:
            res.append(Result(expr, expected, 'ERR', FAIL))
            n_err += 1
            dump += 'ERR: Could not match for result in output.\n'
            dump += f'INPUT: {expr}\n'
            dump += f'OUTPUT:\n{p_out}\n\n'
            continue

        got = search(REAL_REGEX, m[0])[0]

        outcome = ''
        s_ans = got if len(got) <= len(expected) else expected
        l_ans = got if len(got) > len(expected) else expected
        if got == expected:
            outcome = PASS
        elif (len(s_ans) > 1
              and s_ans.find('.') != -1
              and (search(s_ans, l_ans)
                   or search(s_ans[:-1], l_ans[:-1]))):
            outcome = MAYBE
        else:
            outcome = FAIL

        res.append(Result(expr, expected, got, outcome))
    return (res, n_err)


def main(file: TextIOWrapper, prog: str) -> None:
    global dump
    tests = parse_tests(open(file, 'r'))

    if len(tests) == 0:
        print("ERR: No tests to run")
        exit(1)

    res, n_err = run_tests(tests, prog)
    expr_len = max(map(lambda r: len(r.expr), res))
    expect_len = max(map(lambda r: len(r.expected), res))
    got_len = max(map(lambda r: len(r.got), res))
    nr_len = len(str(len(res)))
    bar_len = 35 + expr_len + expect_len + got_len + nr_len

    print('='*bar_len)
    for i, r in enumerate(res):
        print(f'| {i:>{nr_len}}: {r.expr:>{expr_len}} | Expected: {r.expected:>{expect_len}} | Got: {r.got:>{got_len}} | {r.outcome:>{5 + len(RED) + len(RESET)}} |')  # noqa: E501
    print('='*bar_len)

    if n_err:
        rand_str = ''.join([str(randrange(0, 10)) for _ in range(8)])
        dump_file = f'./test-{rand_str}-dump.txt'
        print(f'Writing error outputs to dump file: `{dump_file}`')
        with open(dump_file, 'w') as f:
            f.write(dump)


if __name__ == '__main__':
    ap = ArgParse()
    ap.add_argument(
        'file',
        help='Test file to use.'
    )
    ap.add_argument(
        '-p', '--program',
        default='./lab2',
        help='Executable to run tests on. Defaults to `./lab2`.'
    )
    args = ap.parse_args()
    main(args.file, args.program)
